/* GLCD Library fr's Nokia 6100 
   copyright 2004 Hagen Reddmann
   Resourcenverbrauch ca. 4500 bytes */

#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/signal.h>
#include <avr/delay.h>

#include "glcd.h"

#define glcdDoSwapCoord(x, y) {glcdCoord_t t = x; x = y; y = t;}
#define glcdDoCheckCoord(x, y) {if (x > y) {glcdDoSwapCoord(x, y)}}

//#define USEISR


/* SPI Interrupt handler
   wird nur im Master Mode benutzt und soll sicherstellen das nach einem Mastertransfer das
   Chip Select Signal zurckgesetzt wird.
   USEISR und somit der SPI Interrupt sollte nur benutzt werden falls man mit dem SPI mehrere
   Slaves steuern mu oder wenn man den Stromverbrauch des PCF8833 reduzieren will. Ich habe
   bisher noch nicht getestet wie gro der Stomverbrauch des SPI Interfaces des PCF8833 tatschlich ist.
   Im Falle das der SPI Slave Mode bentigt wird mssensowieso die SPI Routinen angepasst weren. */

#ifdef USEISR
SIGNAL(SIG_SPI) {

    SPCR &= ~(1 << SPE);
    LCD_PORT |= (1 << LCD_CS);
}
#endif

/* Wartet auf Beendigung des Transfers */
inline void glcdDisplayWait(void) {

#ifndef USEISR
    while (SPCR & (1 << SPE)) {
      while (!(SPSR & (1 << SPIF)));
      SPCR &= ~(1 << SPE);
    }
#else
    while (SPCR & (1 << SPE)) ;
    LCD_PORT &= ~(1 << LCD_CS);
#endif
}

void glcdDisplayCommand(const uint8_t cmd) {

    glcdDisplaySend(cmd, 1);
}

void glcdDisplayData(const uint8_t data) {

    glcdDisplaySend(data, 0);
}

/* Transfer von data zum GLCD.
   Der PCF8833 benutzt ein 9Bit Telegram bei dem das MSB angibt ob man Daten oder ein Kommando sendet. */
void glcdDisplaySend(const uint8_t data, const uint8_t cmd) {

    glcdDisplayWait();
    LCD_PORT &= ~(1 << LCD_SCL);
    if (cmd) {LCD_PORT &= ~(1 << LCD_SDA);}
      else {LCD_PORT |= (1 << LCD_SDA);}
    LCD_PORT |= (1 << LCD_SCL);
    SPCR |= (1 << SPE);
    SPDR = data;
}

/* lese bitcount Bits vom GLCD.
   Der PCF8833 empfngt und sendet Daten ber die gleiche MOSI line.
   Mit glcdDisplayRead() knnen auch zB. 33 Bit gelesen werden, allerdings wird dann
   das Bit #32 verworfen. */
uint32_t glcdDisplayRead(uint8_t bitcount) {

	uint32_t data = 0;
    glcdDisplayWait();
    LCD_DDR &= ~(1 << LCD_SDA);
    for (; bitcount > 0; bitcount--) {
      LCD_PORT &= ~(1 << LCD_SCL);
      data += data;
      LCD_PORT |= (1 << LCD_SCL);
      if (LCD_PIN & (1 << LCD_SDA)) {data++;}
    }
    LCD_PORT |= (1 << LCD_CS);               // wir mssen CS zurcksetzen !!
    LCD_DDR |= (1 << LCD_SDA);
#ifndef USEISR
    LCD_PORT &= ~(1 << LCD_CS);
#endif
    return(data);
}

/* sende eine Flash Konstante als Daten ans GLCD, wird hauptschlich zur Einstellung der Farbtabellen bentigt. */
void glcdDisplayConst(uint8_t *buffer, uint8_t count) {

    for (; count > 0; count--) {
      glcdDisplayData(pgm_read_byte(buffer++));
    }
}

void glcdWait(uint16_t ms) {
    for (; ms > 0; ms--) {
      _delay_loop_2(DELAY_OME_MILLISECOND);
    }
}

/* Initialisiere das Display und die GLCD Library */
void glcdDisplayInit(void) {

    static uint8_t __attribute__ ((progmem)) COLOR_RED[] =
      {0x00, 0x03, 0x05, 0x07, 0x09, 0x0B, 0x0D, 0x0F}; // red and green
    static uint8_t __attribute__ ((progmem)) COLOR_BLUE[] =
      {0x00, 0x08, 0x0B, 0x0F}; // blue

    LCD_DDR  |= (1 << LCD_CS) | (1 << LCD_SCL) | (1 << LCD_SDA) | (1 << LCD_RESET);
    LCD_PORT |= (1 << LCD_CS) | (1 << LCD_SCL) | (1 << LCD_SDA);
    SPSR     |= (1 << SPI2X);
#ifndef USEISR
    SPCR      = (1 << MSTR) | (1 << CPOL) | (1 << CPHA);
#else
    SPCR      = (1 << MSTR) | (1 << CPOL) | (1 << CPHA) | (1 << SPIE);
    asm("sei"::);
#endif

// warte min. 65ms, Powerup der MCU,
// auf meinem STK500 gibt es ohne diese Waitstates Problem beim Ein/Ausschalten der Versorgungsspannung !!
    glcdWait(65);

    LCD_PORT &= ~(1 << LCD_RESET);
    asm("nop"::);
    LCD_PORT |=  (1 << LCD_RESET);
    LCD_PORT &= ~(1 << LCD_CS);

    glcdDisplayCommand(SOFT_RESET);
    glcdDisplayCommand(SLEEP_OUT);
    glcdDisplayCommand(BOOSTER_ON);

    glcdDisplayCommand(COLOR_INTERFACE);
    glcdDisplayData(COLOR_8_BIT);
    glcdDisplayCommand(COLOR_SET);
    glcdDisplayConst(COLOR_RED, sizeof COLOR_RED);       // red
    glcdDisplayConst(COLOR_RED, sizeof COLOR_RED);       // green
    glcdDisplayConst(COLOR_BLUE, sizeof COLOR_BLUE);     // blue

    glcdDisplayCommand(MEM_CONTROL);
    glcd_MemCtrl = (1 << MEM_MX) | (0 << MEM_MY) | (1 << MEM_RGB) | (0 << MEM_VW) | (0 << MEM_LAO);
    glcdDisplayData(glcd_MemCtrl);

    glcdSetRect(glcd_Clip, SCREEN_LEFT, SCREEN_TOP, SCREEN_RIGHT, SCREEN_BOTTOM);
    glcdSetRect(glcd_Window, SCREEN_LEFT, SCREEN_TOP, SCREEN_RIGHT, SCREEN_BOTTOM);

    glcdSetBkColor(SCREEN_COLOR);     // Hintergrund
    glcdSetFgColor(BLACK);            // Vordergrund
    glcdSetFrColor(GRAY);             // Rahmen
    glcdSetShColor(SILVER);           // Schatten

    glcdClearScreen();

 // wait for booster
    uint32_t d = 0;
    while (!(d & 0x80000000)) {
  	  glcdDisplayCommand(DISPLAY_STATUS);
  	  d = glcdDisplayRead(33);
    }
    glcdDisplayCommand(DISPLAY_ON);
 // wait for display
    while (!(d & 0x00000400)) {
  	  glcdDisplayCommand(DISPLAY_STATUS);
  	  d = glcdDisplayRead(33);
    }
}

/* setze Display Ausrichtung, 0,90,180,270
   ACHTUNG! bei 90 und 270 werden alle X,Y Koordinaten vertauscht aber nicht die absoluten SCREEN_xxxx Koordinaten.
   Auf einem 132x132 Nokia Display hat das keinerlei Konsequenzen, aber auf einem zB. 176x240 Display wrde das nicht
   funktionieren da dann das Clipping nicht richtig wre. Auf so einem Display mssen die Clipping Rects Clip und Window
   ebenfalls angepasst werden. 
   Das Hardware Scrolling des PCF8833 kann dieses virtuelle drehen der Ausrichtung ebenfalls nicht !!*/
void glcdSetOrientation(const uint8_t orientation) {

  glcdDisplayCommand(MEM_CONTROL);
  glcd_MemCtrl &= ~((1 << MEM_90) | (1 << MEM_VW) | (1 << MEM_MX) | (1 << MEM_MY));
  if (orientation & 1) {
    glcd_MemCtrl |= (1 << MEM_90) | (1 << MEM_VW) | (1 << MEM_MX);
  }
  if (orientation & 2) {
    glcd_MemCtrl ^= (1 << MEM_MY);
  } else {
    glcd_MemCtrl ^= (1 << MEM_MX);
  }
  glcdDisplayData(glcd_MemCtrl);
}

void glcdSetAddr(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2) {

   if (glcd_MemCtrl & (1 << MEM_90)) {
      glcdDoSwapCoord(x1, y1);
      glcdDoSwapCoord(x2, y2);
    }
    glcdDisplayCommand(SET_X_ADDR);
    glcdDisplayData(x1);
    glcdDisplayData(x2);
    glcdDisplayCommand(SET_Y_ADDR);
    glcdDisplayData(y1);
    glcdDisplayData(y2);
    glcdDisplayCommand(MEM_WRITE);
}

void glcdSetPixel(glcdCoord_t x, glcdCoord_t y, const glcdColor_t color) {

    if ((x >= glcd_Clip.X1) & (x <= glcd_Clip.X2) & (y >= glcd_Clip.Y1) & (y <= glcd_Clip.Y2)) {
      glcdSetAddr(x, y, SCREEN_WIDTH -1, SCREEN_HEIGHT -1);
      glcdPutPixel(color);
    }
}

void glcdDoFillRect(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2, const glcdColor_t color) {

    glcdCoord_t x, y;
    if ((x1 > glcd_Clip.X2) | (x2 < glcd_Clip.X1) | (y1 > glcd_Clip.Y2) | (y2 < glcd_Clip.Y1)) {return;}
    if (x1 < glcd_Clip.X1) {x1 = glcd_Clip.X1;}
    if (y1 < glcd_Clip.Y1) {y1 = glcd_Clip.Y1;}
    if (x2 > glcd_Clip.X2) {x2 = glcd_Clip.X2;}
    if (y2 > glcd_Clip.Y2) {y2 = glcd_Clip.Y2;}
    if ((x1 <= x2) & (y1 <= y2)) {
      glcdSetAddr(x1, y1, x2, y2);
      for (y = y1; y <= y2; y++) {
        for (x = x1; x <= x2; x++) {
          glcdPutPixel(color);
        }
      }
    }
}

void glcdFillRect(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2, const glcdColor_t color) {

    glcdDoCheckCoord(x1, x2);
    glcdDoCheckCoord(y1, y2);
    glcdDoFillRect(x1, y1, x2, y2, color);
}

void glcdRectangle(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2) {

    glcdDoCheckCoord(x1, x2);
    glcdDoCheckCoord(y1, y2);
    if (glcdFgColor != NONE) {
      glcdDoFillRect(x1, y1, x2, y1, glcdFgColor);
      glcdDoFillRect(x2, y1, x2, y2, glcdFgColor);
      glcdDoFillRect(x1, y2, x2, y2, glcdFgColor);
      glcdDoFillRect(x1, y1, x1, y2, glcdFgColor);
    }
    if (glcdBkColor != NONE) {
      glcdDoFillRect(++x1, ++y1, --x2, --y2, glcdBkColor);
    }
}

void glcdFrame(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2) {

    glcdDoCheckCoord(x1, x2);
    glcdDoCheckCoord(y1, y2);

    if (glcdShColor != NONE) {
      glcdDoFillRect(x1 + 4, y2, x2 - 2, y2, glcdShColor);
      glcdDoFillRect(x2, y1 +4, x2, y2 -2, glcdShColor);
      x2--;
      y2--;
      glcdDoFillRect(x2 -2, y2 -2, x2, y2, glcdShColor);
    }
    if (glcdFrColor != NONE) {
      glcdDoFillRect(x1 +2, y1, x2 -2, y1, glcdFrColor);
      glcdDoFillRect(x1 +2, y2, x2 -2, y2, glcdFrColor);
      glcdDoFillRect(x2, y1 +2, x2, y2 -2, glcdFrColor);
      glcdDoFillRect(x1, y1 +2, x1, y2 -2, glcdFrColor);
      x1++;
      y1++;
      x2--;
      y2--;
      glcdSetPixel(x1, y1, glcdFrColor);
      glcdSetPixel(x2, y1, glcdFrColor);
      glcdSetPixel(x1, y2, glcdFrColor);
      glcdSetPixel(x2, y2, glcdFrColor);
    }
    if (glcdBkColor != NONE) {
      glcdDoFillRect(x1, y1 +1, x1, y2 -1, glcdBkColor);
      glcdDoFillRect(x2, y1 +1, x2, y2 -1, glcdBkColor);
      glcdDoFillRect(x1 +1, y1, x2 -1, y2, glcdBkColor);
    }
}

void glcdLine(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2) {

    if (glcdFgColor != NONE) {
      if ((x1 == x2) | (y1 == y2)) {
        glcdFillRect(x1, y1, x2, y2, glcdFgColor);
      } else {

        glcdCoord_t dx, dy, sx, sy;

        if (x1 < x2) {dx = x2 - x1; sx = +1;}
          else {dx = x1 - x2; sx = -1;}
        if (y1 < y2) {dy = y2 - y1; sy = +1;}
          else {dy = y1 - y2; sy = -1;}

        uint8_t swap = (dx > dy) ? 0 : 1;
        if (swap) {
          glcdDoSwapCoord(x1, y1);
          glcdDoSwapCoord(dx, dy);
          glcdDoSwapCoord(sx, sy);
        }

        int16_t e = dy - dx + dy;
        glcdCoord_t x3 = x1;
        glcdCoord_t y3 = y1;
        x2 = x1;
        y2 = y1;
/* Diese Loop ist suboptimal. Man knnte mehr rausholen wenn man die Hardware besser ausnutzen wrde, aber
   dann wre diese Routine eben nicht mehr Hardwareunabhngig so wie jetzt. */
        for (glcdCoord_t i = 0; i <= dx; i++) {
          if ((x1 != x2) & (y1 != y2)) {
            if (swap) {
              glcdDoSwapCoord(x2, y2);
              glcdDoSwapCoord(x3, y3);
            }
            glcdDoCheckCoord(x2, x3);
            glcdDoCheckCoord(y2, y3);
            glcdDoFillRect(x2, y2, x3, y3, glcdFgColor);
            x2 = x1;
            y2 = y1;
          };
          x3 = x1;
          y3 = y1;
          while (e >= 0) {
            y1 += sy;
            e  -= dx;
            e  -= dx;
          }
          x1 += sx;
          e  += dy;
          e  += dy;
        }
        if (swap) {
          glcdDoSwapCoord(x2, y2);
          glcdDoSwapCoord(x3, y3);
        }
        glcdDoCheckCoord(x2, x3);
        glcdDoCheckCoord(y2, y3);
        glcdDoFillRect(x2, y2, x3, y3, glcdFgColor);
      }
    }
}

/* Hilfsfunktion fr Ellipsen */
void glcdDoPixelLine(int16_t x1, uint16_t x2, const int16_t y, const uint8_t fill) {

    if ((y >= glcd_Clip.Y1 ) & (y <= glcd_Clip.Y2)) {
      if (x1 < glcd_Clip.X1) {x1 = glcd_Clip.X1;} else {
        if (x1 > glcd_Clip.X2) {return;} else {
          if (glcdFgColor != NONE) {
            glcdSetPixel(x1, y, glcdFgColor);
          }
          x1++;
        }
      }
      if (x2 < glcd_Clip.X1) {return;} else {
        if (x2 > glcd_Clip.X2) {x2 = glcd_Clip.X2;} else {
          if (glcdFgColor != NONE) {
            glcdSetPixel(x2, y, glcdFgColor);
          }
          x2--;
        }
      }
      if ((fill) & (glcdBkColor != NONE) & (x1 <= x2)) {
        glcdDoFillRect(x1, y, x2, y, glcdBkColor);
      }
    }
}

void glcdCircle(glcdCoord_t x, glcdCoord_t y, glcdCoord_t r) {

    glcdEllipse(x, y, r, r);
}

void glcdEllipse(glcdCoord_t x, glcdCoord_t y, glcdCoord_t a, glcdCoord_t b) {

   if ((a == 0) | (b == 0)) {return;}
   if ((a > 180) | (b > 180)) {
     if (glcdBkColor != NONE) {
       glcdFillRect(0, 0, SCREEN_WIDTH -1, SCREEN_HEIGHT -1, glcdBkColor);
     }
     return;
   }

   glcdCoord2_t aa = a * a;
   glcdCoord2_t bb = b * b;
   int32_t er, cr, ir;
   int16_t ys,ye,xs,xe;

   cr = bb >> 1;
   cr = cr * (a + a -1);
   ir = aa >> 1;
   ir = -ir;
   er = 0;

   xs = x;
   xs = xs - a;
   xe = x;
   xe = xe + a;
   ys = y;
   ye = y;
   while (cr >= ir) {
     glcdDoPixelLine(xs, xe, ys, 1);
     if (ys != ye) {
       glcdDoPixelLine(xs, xe, ye, 1);
     }
     ys--;
     ye++;
     ir += aa;
     er += ir;
     if (2 * er > cr) {
       er -= cr;
       cr -= bb;
       xs++;
       xe--;
     }
   }

   cr = aa >> 1;
   cr = cr * (b + b -1);
   ir = bb >> 1;
   ir = -ir;
   er = 0;

   xs = x;
   xe = x;
   ys = y;
   ys = ys - b;
   ye = y;
   ye = ye + b;
   uint8_t fill = 1;
   while (ir <= cr) {
     glcdDoPixelLine(xs, xe, ys, fill);
     if (ys != ye) {
       glcdDoPixelLine(xs, xe, ye, fill);
     }
     fill = 0;
     ir += bb;
     er += ir;
     if (2 * er > cr) {
       er -= cr;
       cr -= aa;
       ys++;
       ye--;
       fill = 1;
     }
     xs--;
     xe++;
   }
}

void glcdSelectFont(PGM_VOID_P data, glcdFontFunc_t callback) {

    glcd_FontData = (glcdFontData_t)(data);
    glcd_FontRead = callback;
    glcd_FontWidth = glcd_FontRead(glcd_FontData + 2);
    glcd_FontHeight = glcd_FontRead(glcd_FontData + 3);
    glcd_FontBitsPixel = glcd_FontRead(glcd_FontData + 4);
    glcd_FontFirstChar = glcd_FontRead(glcd_FontData + 5);
    glcd_FontLastChar = glcd_FontRead(glcd_FontData + 6);
}

uint8_t glcdFontRead(const glcdFontData_t index) {
    
    return(pgm_read_byte(index));
}

void glcdNewLine(void) {

    glcd_Cursor.X = glcd_Window.X1;
    glcd_Cursor.Y += glcd_FontHeight;
    if (glcd_Cursor.Y + glcd_FontHeight > glcd_Window.Y2) {
   	  glcd_Cursor.Y = glcd_Window.Y1;
    }
}

int glcdDrawChar(char c) {

#define HEADER_SIZE  7

    if (c == '\n') {
      glcdNewLine();
      return(0);
    }
  // ist das Zeichen im Font definiert und auch ein Font selektiert ??
    if ((c < glcd_FontFirstChar) | (c > glcd_FontLastChar) | (glcd_FontData == 0)) {
      return(0);
    }

  // ist das Zeichen im Font deklariert aber nicht definiert ??
    uint8_t charwidth = glcd_FontRead(glcd_FontData + HEADER_SIZE + c - glcd_FontFirstChar);
    if (charwidth == 0) {return(0);}

  // alle Zeichen kleiner #128 haben rechts eine implizite Leerspalte
    uint8_t width = charwidth;
    if (c < 128) {width++;}

  // wenn wir den Zeichensatz mit fester Breite zeichnen sollen dann setze die Width entsprechend
    if ((glcd_Flags.FixedFont == 1) & (width <= glcd_FontWidth)) {
      width = glcd_FontWidth;
    }

  // passt das Zeichen noch in das Zeichen-Fenster ?? falls nein eventuell ein Linefeed einfgen
    if ((glcd_Flags.AutoLineFeed == 1) & (glcd_Cursor.X + width >= glcd_Window.X2)) {
      glcdNewLine();
    }

  // ist das Zeichen berhaupt noch ansatzweise sichtbar ?
    if ((glcd_Cursor.X > glcd_Clip.X2) | (glcd_Cursor.Y > glcd_Clip.Y2)) {
      return(0);
    }

    glcdCoord_t x = glcd_Cursor.X;
    glcdCoord_t y = glcd_Cursor.Y;
    glcd_Cursor.X += width;

  // ist das Zeichen berhaupt noch ansatzweise sichtbar ?
    if ((x < glcd_Clip.X1) | (y + glcd_FontHeight < glcd_Clip.Y1)) {
      return(0);
    }

    uint8_t i = 0;

  // falls FixedFont mssen die Zeichen innerhalb glcd_FontWidth zentriert werden, aber nur
  // wenn das Zeichen schmaler als glcd_FontWidth ist. Wir fllen dann auch gleich den rechten
  // und linken Rand des Zeichens, ebenfalls nur wenn die Hintergrundfarbe nicht transparent ist.
    if ((glcd_Flags.FixedFont == 1) & (width > charwidth)) {
      i = (width - charwidth) / 2;
      if (glcdBkColor != NONE) {
        glcdDoFillRect(x, y, x + i -1, y + glcd_FontHeight -1, glcdBkColor);
      }
      x += i;
    }
    i = width - charwidth - i;
    if ((i > 0) & (glcdBkColor != NONE)) {
      glcdDoFillRect(x + charwidth, y, x + charwidth + i -1, y + glcd_FontHeight -1, glcdBkColor);
    }

  // mssen die eigentlichen Zeichendaten gezeichnet werden ??
    if ((x > glcd_Clip.X2) | (x + charwidth < glcd_Clip.X1)) {
      return(width);
    }

  // berechne nun den Bitindex zu den Bitdaten im Font
    glcdFontData_t data = glcd_FontData + HEADER_SIZE;
    uint32_t index = 0;

    for (i = glcd_FontFirstChar; i < c; i++) {
      index += glcd_FontRead(data++);
    }
    index = (index * glcd_FontHeight) << (glcd_FontBitsPixel -1);

  // positioniere data auf's erste Byte mit Fontdaten zum Zeichen und hole die ersten Bits
    data = glcd_FontData + HEADER_SIZE + glcd_FontLastChar - glcd_FontFirstChar +1 + index / 8;
    uint8_t bitsindex = index % 8;
    uint8_t bits = glcd_FontRead(data++) >> bitsindex;
    uint8_t mask = 0xFF >> (8 - glcd_FontBitsPixel);
    uint8_t invalid = 1;
    glcdColor_t color;

  // wir senden die Bitdaten der einzelnen Pixel jeweils Spaltenweise in Y Richtung
    glcdDisplayCommand(MEM_CONTROL);
    glcd_MemCtrl ^= (1 << MEM_VW);
    glcdDisplayData(glcd_MemCtrl);

  // Zeichne nun die charwidth Spalten des Zeichens, bercksichtige das Clipping.
  // Diese Loop knnte das Clipping schon in advance durchfhren um eine eventuell
  // besssere Performance zu bringen, zZ. belassen wir es aber erstmal.
    for (; charwidth > 0; charwidth--) {
      for (y = glcd_Cursor.Y; y < glcd_Cursor.Y + glcd_FontHeight; y++) {
        if ((x >= glcd_Clip.X1) & (x <= glcd_Clip.X2) & (y >= glcd_Clip.Y1) & (y <= glcd_Clip.Y2)) {
  // falls der Font in der Farbanzahl untersttzt wird benutze Farbtabelle ansonsten 1 zu 1
          if (glcd_FontBitsPixel <= COLOR_TABLE_BITS) {
            color = glcd_Colors[bits & mask];
          } else {color = bits & mask;}
          if (color != NONE) {
            if (invalid) {
              glcdSetAddr(x, y, glcd_Cursor.X, glcd_Cursor.Y + glcd_FontHeight -1);
              if (y == glcd_Cursor.Y) {
                invalid = 0;
              }
            }
            glcdPutPixel(color);
          } else {
            invalid = 1;
          }
        } else {
          invalid = 1;
        }

        bits = bits >> glcd_FontBitsPixel;
        bitsindex += glcd_FontBitsPixel;
        if (bitsindex == 8) {
          bits = glcd_FontRead(data++);
          bitsindex = 0;
        }
      }
      x++;
    }

    glcdDisplayCommand(MEM_CONTROL);
    glcd_MemCtrl ^= (1 << MEM_VW);
    glcdDisplayData(glcd_MemCtrl);

    return(width);
}

void glcdPrint(char *c) {

    while (*c != 0) {
      glcdDrawChar(*c++);
    }
}

int glcdCharWidth(char c) {

    if ((c < glcd_FontFirstChar) | (c > glcd_FontLastChar) | (glcd_FontData == 0)) {
      return(0); // character not defined
    } else {
      int i = glcd_FontRead(glcd_FontData + HEADER_SIZE + c - glcd_FontFirstChar);
      if (c < 128) {i++;}
      return(i);
    }
}

int glcdCharsWidth(char *c) {

    int i = 0;
    while (*c != 0) {
      i += glcdCharWidth(*c++);
    }
    return(i);
}
