/* GLCD Library fr's Nokia 6100 
   copyright 2004 Hagen Reddmann */

#include <avr/io.h>
#include <avr/pgmspace.h>

#ifndef GLCD_H

#define GLCD_H

#if defined (__AVR_ATmega8__)
  #define LCD_PORT                PORTB
  #define LCD_PIN                 PINB
  #define LCD_DDR                 DDRB
  #define LCD_CS                  PB2   // SS
  #define LCD_SDA                 PB3   // MOSI
  #define LCD_RESET               PB4   // MISO
  #define LCD_SCL                 PB5   // SCK
#elif defined (__AVR_ATmega32__)
  #define LCD_PORT                PORTB
  #define LCD_PIN                 PINB
  #define LCD_DDR                 DDRB
  #define LCD_CS                  PB4   // SS
  #define LCD_SDA                 PB5   // MOSI
  #define LCD_RESET               PB6   // MISO
  #define LCD_SCL                 PB7   // SCK
#else
  #error Please define in glcd.h the hardware pins for your device !
#endif

#define DELAY_OME_MILLISECOND   4000  // 1 millisecond at 16Mhz

#define SCREEN_WIDTH            132   // Nokia6100, wobei aber nur 1,1-130,130 sichtbar sind. Denoch bentigen wir vollen
#define SCREEN_HEIGHT           132   // Zugriff auf den RAM wenn wir zB. das Scrolling nutzen wollen
#define SCREEN_LEFT             1
#define SCREEN_TOP              1
#define SCREEN_RIGHT            (SCREEN_WIDTH -2)
#define SCREEN_BOTTOM           (SCREEN_HEIGHT -2)

#define SCREEN_COLOR            WHITE // sollte gendert werden fr andere Hintergrundfarbe
#define COLOR_TABLE_BITS        2     // 2^2 = 4 Farben in Farbtabelle, wichtig fr Fonts

#define RGB(r,g,b) ((r & 0xE0) | ((g & 0xE0) >> 3) | (b >> 6))

#define NONE                    RGB(0x00, 0x20, 0x00)  // == tansparent, ein grn Ton weniger !!
#define BLACK                   RGB(0x00, 0x00, 0x00)  // fr 8 Bit Farben R,G,B  3,3,2 Bits
#define WHITE                   RGB(0xFF, 0xFF, 0xFF)
#define RED                     RGB(0xFF, 0x00, 0x00)
#define GREEN                   RGB(0x00, 0xFF, 0x00)
#define BLUE                    RGB(0x00, 0x00, 0xFF)
#define YELLOW                  RGB(0xFF, 0xFF, 0x00)
#define MAGENTA                 RGB(0xFF, 0x00, 0xFF)
#define CYAN                    RGB(0x00, 0xFF, 0xFF)
#define GRAY                    RGB(0x80, 0x80, 0x40)
#define SILVER                  RGB(0xA0, 0xA0, 0x80)
#define GOLD                    RGB(0xA0, 0xA0, 0x40)


typedef uint8_t glcdCoord_t;          // Koordinaten
typedef uint16_t glcdCoord2_t;        // intern zur Koordinatenberechnung
typedef uint8_t glcdColor_t;          // Farbe, fr 256 Farbe uint8_t
typedef uint8_t *glcdFontData_t;      // zeiger auf Fontdaten
typedef uint8_t (*glcdFontFunc_t)(const glcdFontData_t index); // Fontdaten-Lese-Callback, somit knnen die Fontdaten extern gespeichert sein

typedef struct {
    glcdCoord_t X;
    glcdCoord_t Y;
} glcdPoint_t;

typedef struct {
    glcdCoord_t X1;
    glcdCoord_t Y1;
    glcdCoord_t X2;
    glcdCoord_t Y2;
} glcdRect_t;

typedef struct {
    uint8_t AutoLineFeed:1,        // Fontroutine nutzt einen impliziten Zeilenumbruch innerhalb von glcd_Window
            FixedFont:1;           // Fontroutine soll Font mit fester Zeichenbreite darstellen, ansonsten proportional
} glcdFlags_t;



glcdColor_t     glcd_Colors[1 << COLOR_TABLE_BITS]; // Farbtabelle, Colors[0] = Hintergrund = Fllfarbe, Color[1] = Vordergrund = Linienfarbe
uint8_t         glcd_MemCtrl;
glcdFlags_t     glcd_Flags;
glcdPoint_t     glcd_Cursor;          // aktuelle Position fr Schriften
glcdRect_t      glcd_Window;          // aktuelles Fenster fr Textausgaben
glcdRect_t      glcd_Clip;            // Clipping Koordinaten, alles was ausserhalb ist wird nicht dargestellt
char            glcd_FontFirstChar;   // erstes Zeichen im Font
char            glcd_FontLastChar;    // letztes Zeichen im Font
uint8_t         glcd_FontWidth;       // fixed Zeichenbreite des Fonts
uint8_t         glcd_FontHeight;      // Zeichenhhe
uint8_t         glcd_FontBitsPixel;   // Farbtiefe des Fonts
glcdFontFunc_t  glcd_FontRead;        // User definierte Readcallback fr die Fontdaten
glcdFontData_t  glcd_FontData;        // die eigentlichen Fontdaten beginnen hier, abhngig von der Readcallback

// todo: Fllen mit Brush, Linien mit Pen

void glcdDisplayInit(void);                                   // initialize GLCD
void glcdDisplayCommand(const uint8_t cmd);                   // send GLCD 9bit command
void glcdDisplayData(const uint8_t data);                     // send GLCD 9bit data
void glcdDisplaySend(const uint8_t data, const uint8_t cmd);  // transfer to GLCD
uint32_t glcdDisplayRead(uint8_t bitcount);                   // read bitcount bits from GLCD
void glcdWait(uint16_t ms);

void glcdSetOrientation(const uint8_t orientation);           // Display Orientation {0,1,2,3} -> {0, 90, 180, 270}
void glcdSetAddr(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2);
void glcdSetPixel(glcdCoord_t x, glcdCoord_t y, const glcdColor_t color);
void glcdFillRect(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2, const glcdColor_t color);
void glcdRectangle(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2);
void glcdFrame(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2);
void glcdLine(glcdCoord_t x1, glcdCoord_t y1, glcdCoord_t x2, glcdCoord_t y2);
void glcdCircle(glcdCoord_t x, glcdCoord_t y, glcdCoord_t r);
void glcdEllipse(glcdCoord_t x, glcdCoord_t y, glcdCoord_t a, glcdCoord_t b);

// Font routinen, can be used as Icons
void glcdSelectFont(PGM_VOID_P data, glcdFontFunc_t callback);  // select given font and read above font vaiables in
void glcdNewLine(void);                                         // linefeed in glcd_Window bounds
int  glcdDrawChar(char c);                                      // paint single char
void glcdPrint(char *c);                                        // paint string
int  glcdCharWidth(char c);                                     // compute pixel width of single char
int  glcdCharsWidth(char *c);                                   // compute pixel width of string
uint8_t glcdFontRead(const glcdFontData_t index);               // possible Font-Read-Callback for reading from Flash     

// todo:
// void glcdBezier(?????);

#define glcdBkColor glcd_Colors[0]
#define glcdFgColor glcd_Colors[1]
#define glcdFrColor glcd_Colors[2]
#define glcdShColor glcd_Colors[3]
#define glcdMoveTo(x, y) {glcd_Cursor.X = x; glcd_Cursor.Y = y;}
#define glcdSetColor(index, color) {glcd_Colors[index] = color;}
#define glcdSetBkColor(color) {glcdSetColor(0, color);}
#define glcdSetFgColor(color) {glcdSetColor(1, color);}
#define glcdSetFrColor(color) {glcdSetColor(2, color);}
#define glcdSetShColor(color) {glcdSetColor(3, color);}
#define glcdSetColors(fgColor, bkColor) {glcdSetFgColor(fgColor); glcdSetBkColor(bkColor);}
#define glcdPutPixel(color) {glcdDisplayData(color);}
#define glcdClearScreen() {glcdFillRect(0, 0, SCREEN_WIDTH -1, SCREEN_HEIGHT -1, SCREEN_COLOR);}
#define glcdSetRect(r,x1,y1,x2,y2) {r.X1 = x1; r.Y1 = y1; r.X2 = x2; r.Y2 = y2;}

// Commands

#define NO_OPERATION            0x00
#define SOFT_RESET              0x01
#define BOOSTER_OFF             0x02
#define BOOSTER_ON              0x03
#define DISPLAY_INFO            0x04   // read 25bits where MSB are dummy trash
#define DISPLAY_STATUS          0x09   // read 33bits where MSB are dummy trash
#define SLEEP_IN                0x10
#define SLEEP_OUT               0x11
#define PARTIAL_MODE_ON         0x12
#define NORMAL_MODE_ON          0x13
#define INVERSION_OFF           0x20
#define INVERSION_ON            0x21
#define PIXELS_OFF              0x22
#define PIXELS_ON               0x23
#define SET_CONTRAST            0x25
  // CONTRAST_VALE              0xXX
#define DISPLAY_OFF             0x28
#define DISPLAY_ON              0x29
#define SET_X_ADDR              0x2A  // ATENTION! when X,Y address are changed BOTH must be changed
  // START_X                    0xXX
  // STOP_X                     0xXX
#define SET_Y_ADDR              0x2B
  // START_Y                    0xXX
  // STOP_Y                     0xXX
#define MEM_WRITE               0x2C
  // x times pixel data         0xXX
#define COLOR_SET               0x2D
  // 8 times red palette        0xXX
  // 8 times green palette      0xXX
  // 4 times blue palette       0xXX
#define PARTIAL_AREA            0x30
  // START_Y                    0xXX
  // STOP_Y                     0xXX
#define SCROLL_DEFINITON        0x33  // always vertical
  // TOP_FIXED_LINES            0xXX
  // SCROLL_LINES               0xXX
  // BOTTOM_FIXED_LINES         0xXX
#define SET_SCROLL              0x37  // SEP scroll entry point
  // ROW_Y                      0xxx
#define MEM_CONTROL             0x36
  // MEM_FLAGS                  0xXX
  #define MEM_90                0     // virtual flag not for PCF8833
  #define MEM_RGB               3     // color order
  #define MEM_LAO               4     // line access order
  #define MEM_VW                5     // vertical write
  #define MEM_MX                6     // mirror X
  #define MEM_MY                7     // mirror Y
#define IDLE_OFF                0x38
#define IDLE_ON                 0x39
#define COLOR_INTERFACE         0x3A
  // pixelformat                0xXX
  #define COLOR_8_BIT           0x02
  #define COLOR_12_BIT          0x03
  #define COLOR_16_BIT          0x05
#define SET_VOP                 0xB0  // irrelevant for Nokia6100
  // 2x                         0xXX
#define BOTTOM_ROW_SWAP_OFF     0xB4
#define BOTTOM_ROW_SWAP_ON      0xB5
#define TOP_ROW_SWAP_OFF        0xB6
#define TOP_ROW_SWAP_ON         0xB7
#define SUPER_FRAME_INV_OFF     0xB8
#define SUPER_FRAME_INV_ON      0xB9
#define DATA_ORDER_MSB          0xBA
#define DATA_ORDER_LSB          0xBB
#define TEMP_READ               0xC8  // read
  // temp                       0xXX


#endif
